#include "skype_api.h"
#include <conio.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <rpcdce.h>

#define eFree free
#define eMalloc malloc

#ifndef bool
#define bool BOOL
#endif
#define true TRUE
#define false FALSE
struct SKYPE;
typedef struct SKYPE SKYPE;
struct SKYPE {
	HWND hwnd;
	UINT discover;
	UINT attach;
	HANDLE console_input_thread;
	HANDLE volatile console;
	DWORD console_mode;
	HWND SkypeAPIWindow;
	CRITICAL_SECTION skypeAPIWindowCriticalSection;
	ATTACHE_LISTENER al;
	DATA_LISTENER dl;
};
SKYPE *gskype;
LRESULT connect_skype(SKYPE *skype);
bool Initialize_CreateWindowClass(HINSTANCE hInstance,
		char *acInit_WindowClassName, int class_name_size);
static LRESULT APIENTRY SkypeAPITest_Windows_WindowProc(HWND hWindow,
		UINT uiMessage, WPARAM uiParam, LPARAM ulParam);
HWND Initialize_CreateMainWindow(HINSTANCE hInstance,
		const char *acInit_WindowClassName);
DWORD WINAPI InputProcessingThread(LPVOID lpParam);
char acInit_WindowClassName[128];

BOOL open_skype_api(HINSTANCE hInstance,BOOL debug_console,ATTACHE_LISTENER al,DATA_LISTENER dl){
	UINT uiGlobal_MsgID_SkypeControlAPIAttach;
	UINT uiGlobal_MsgID_SkypeControlAPIDiscover;
	HWND hInit_MainWindowHandle;
	SKYPE *skype;
	uiGlobal_MsgID_SkypeControlAPIAttach = RegisterWindowMessage(
			"SkypeControlAPIAttach");
	if (uiGlobal_MsgID_SkypeControlAPIAttach == 0)
		return FALSE;

	uiGlobal_MsgID_SkypeControlAPIDiscover = RegisterWindowMessage(
			"SkypeControlAPIDiscover");

	if (uiGlobal_MsgID_SkypeControlAPIDiscover == 0)
		return FALSE;
	if (!Initialize_CreateWindowClass(hInstance, acInit_WindowClassName, 128))
		return FALSE;

	hInit_MainWindowHandle = Initialize_CreateMainWindow(hInstance,
			acInit_WindowClassName);

	if (!hInit_MainWindowHandle) {
		UnregisterClass(acInit_WindowClassName, hInstance);
		return FALSE;
	}

	skype = eMalloc(sizeof(SKYPE));
	skype->hwnd = hInit_MainWindowHandle;
	skype->discover = uiGlobal_MsgID_SkypeControlAPIDiscover;
	skype->attach = uiGlobal_MsgID_SkypeControlAPIAttach;
	if (debug_console) {
		HANDLE hThread = CreateThread(NULL,
				0,
				InputProcessingThread,
				skype,
				0,
				NULL);
		skype->console_input_thread = hThread;
	} else {
		skype->console_input_thread = NULL;
	}
	gskype = skype;
	skype->al = al;
	skype->dl = dl;
	InitializeCriticalSection(&skype->skypeAPIWindowCriticalSection);
	connect_skype(skype);
	return TRUE;
}

void free_skype(SKYPE *skype) {
	if (skype->console != (HANDLE) 0) {
		SetConsoleMode(skype->console, skype->console_mode);
		CloseHandle(skype->console);
		skype->console = (HANDLE) 0;
		WaitForSingleObject(skype->console_input_thread, INFINITE);
	}
	eFree(skype);
}

bool Global_Console_ReadRow(char *pacPromptBuffer, unsigned int uiMaxLength,
		SKYPE *skype) {
	HANDLE hConsoleHandle, hDuplicatedConsoleHandle;
	DWORD ulCharactersRead, ulConsoleMode;
	unsigned int uiNewLength;
	BOOL fReadConsoleResult;
	bool fReturnStatus;
	char cCharacter;

	fReturnStatus = false;
	while ((hConsoleHandle = GetStdHandle(STD_INPUT_HANDLE))
			!=INVALID_HANDLE_VALUE) {
		if (DuplicateHandle(GetCurrentProcess(), hConsoleHandle,
				GetCurrentProcess(), &hDuplicatedConsoleHandle, 0, FALSE,
				DUPLICATE_SAME_ACCESS) == FALSE)
			break;
		GetConsoleMode(hDuplicatedConsoleHandle, &ulConsoleMode);
		SetConsoleMode(hDuplicatedConsoleHandle, ENABLE_LINE_INPUT
				| ENABLE_PROCESSED_INPUT | ENABLE_ECHO_INPUT);

		skype->console = hDuplicatedConsoleHandle;
		skype->console_mode = ulConsoleMode;

		fReadConsoleResult = ReadConsole(skype->console,
				(LPVOID) pacPromptBuffer, uiMaxLength, &ulCharactersRead,
				(LPVOID) 0);

		if (skype->console == (HANDLE) 0)
			break;

		skype->console = (HANDLE) 0;
		SetConsoleMode(hDuplicatedConsoleHandle, ulConsoleMode);
		CloseHandle(hDuplicatedConsoleHandle);
		if (fReadConsoleResult == FALSE || ulCharactersRead > uiMaxLength)
			break;
		pacPromptBuffer[ulCharactersRead] = 0;
		uiNewLength = ulCharactersRead;
		while (uiNewLength != 0) {
			cCharacter = pacPromptBuffer[uiNewLength - 1];
			if (cCharacter != '\r' && cCharacter != '\n')
				break;
			uiNewLength--;
		}
		pacPromptBuffer[uiNewLength] = 0;
		fReturnStatus = true;
		break;
	}
	if (fReturnStatus == false)
		pacPromptBuffer[0] = 0;
	return (fReturnStatus);
}
DWORD WINAPI InputProcessingThread(LPVOID lpParam) {
	SKYPE *skype = (SKYPE *) lpParam;
	static char acInputRow[1024];
	bool fProcessed;
	while (Global_Console_ReadRow(acInputRow, sizeof(acInputRow) - 1, skype)) {
		if (stricmp(acInputRow, "#quit") == 0 || stricmp(acInputRow, "#exit")
				== 0)
			break;
		fProcessed = false;
		if (stricmp(acInputRow, "#connect") == 0) {
			SendMessage(HWND_BROADCAST, skype->discover, (WPARAM) skype->hwnd,
					0);
			fProcessed = true;
		}
		if (stricmp(acInputRow, "#disconnect") == 0) {
			EnterCriticalSection(&skype->skypeAPIWindowCriticalSection);
			skype->SkypeAPIWindow = NULL;
			LeaveCriticalSection(&skype->skypeAPIWindowCriticalSection);
			printf("!!! Disconnected\n");
			fProcessed = true;
		}
		if (fProcessed == false && skype->SkypeAPIWindow != NULL) {
			COPYDATASTRUCT oCopyData;
			// send command to skype
			oCopyData.dwData = 0;
			oCopyData.lpData = acInputRow;
			oCopyData.cbData = strlen(acInputRow) + 1;
			if (oCopyData.cbData != 1) {
				if (SendMessage(skype->SkypeAPIWindow, WM_COPYDATA,
						(WPARAM) skype->hwnd, (LPARAM) &oCopyData) == FALSE) {
					EnterCriticalSection(&skype->skypeAPIWindowCriticalSection);
					skype->SkypeAPIWindow = NULL;
					LeaveCriticalSection(&skype->skypeAPIWindowCriticalSection);
					printf("!!! Disconnected\n");
					if(skype->al){
						skype->al(DISCONNECTED);
					}
				}
			}
		}
	}
	PostMessage(skype->hwnd, WM_CLOSE, 0, 0);
	return 0;
}
void send_cmd_i(const char* cmd,int cmd_len){
	SKYPE *skype = gskype;
	COPYDATASTRUCT oCopyData;
	oCopyData.dwData = 0;
	oCopyData.lpData = (PVOID)cmd;
	if(cmd_len>1)
		oCopyData.cbData = cmd_len;
	else
		oCopyData.cbData = strlen(cmd) + 1;
	if (oCopyData.cbData != 1) {
			if (SendMessage(skype->SkypeAPIWindow, WM_COPYDATA,
					(WPARAM) skype->hwnd, (LPARAM) &oCopyData) == FALSE) {
				EnterCriticalSection(&skype->skypeAPIWindowCriticalSection);
					skype->SkypeAPIWindow = NULL;
					LeaveCriticalSection(&skype->skypeAPIWindowCriticalSection);
					if(skype->al){
						skype->al(DISCONNECTED);
					}
			}
	}
}
LRESULT connect_skype(SKYPE *skype) {
	return PostMessage(HWND_BROADCAST, skype->discover, (WPARAM) skype->hwnd, 0);
}

HWND Initialize_CreateMainWindow(HINSTANCE hInstance,
		const char *acInit_WindowClassName) {
	HWND hwnd = CreateWindowEx(WS_EX_APPWINDOW | WS_EX_WINDOWEDGE,
			acInit_WindowClassName, "",
			WS_BORDER | WS_SYSMENU | WS_MINIMIZEBOX, CW_USEDEFAULT,
			CW_USEDEFAULT, 128, 128, NULL, 0, hInstance, 0);
	return hwnd;
}

bool Initialize_CreateWindowClass(HINSTANCE hInstance,
		char *acInit_WindowClassName, int class_name_size) {
	unsigned char *paucUUIDString;
	RPC_STATUS lUUIDResult;
	UUID oUUID;

	lUUIDResult = UuidCreate(&oUUID);
	if (lUUIDResult != RPC_S_OK && lUUIDResult != RPC_S_UUID_LOCAL_ONLY) {
		return FALSE;
	}
	if (UuidToString(&oUUID, &paucUUIDString) != RPC_S_OK) {
		return FALSE;
	}
	strcpy(acInit_WindowClassName, "Skype-API-");
	strcat(acInit_WindowClassName, (char *) paucUUIDString);
	RpcStringFree(&paucUUIDString);

	WNDCLASS oWindowClass;
	oWindowClass.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
	oWindowClass.lpfnWndProc = (WNDPROC) &SkypeAPITest_Windows_WindowProc;
	oWindowClass.cbClsExtra = 0;
	oWindowClass.cbWndExtra = 0;
	oWindowClass.hInstance = hInstance;
	oWindowClass.hIcon = NULL;
	oWindowClass.hCursor = NULL;
	oWindowClass.hbrBackground = NULL;
	oWindowClass.lpszMenuName = NULL;
	oWindowClass.lpszClassName = acInit_WindowClassName;
	return RegisterClass(&oWindowClass);
}

static LRESULT APIENTRY SkypeAPITest_Windows_WindowProc(HWND hWindow,
		UINT uiMessage, WPARAM uiParam, LPARAM ulParam) {
	LRESULT lReturnCode;
	bool fIssueDefProc;
	lReturnCode = 0;
	fIssueDefProc = false;
	SKYPE *skype = gskype;

	switch (uiMessage) {
	case WM_CREATE:
		break;
	case WM_DESTROY:
		skype->hwnd = NULL;
		PostQuitMessage(0);
		break;
	case WM_COPYDATA:
		if (skype->SkypeAPIWindow == (HWND) uiParam) {
			PCOPYDATASTRUCT poCopyData = (PCOPYDATASTRUCT) ulParam;
			if (skype->dl) {
				skype->dl(poCopyData->lpData, poCopyData->cbData);
			}
			lReturnCode = 1;
		}
		break;
	default:
		if (skype && uiMessage == skype->attach) {
			switch (ulParam) {
			case SKYPECONTROLAPI_ATTACH_SUCCESS:
				EnterCriticalSection(&skype->skypeAPIWindowCriticalSection);
				skype->SkypeAPIWindow = (HWND) uiParam;
				LeaveCriticalSection(&skype->skypeAPIWindowCriticalSection);
				break;
			case SKYPECONTROLAPI_ATTACH_PENDING_AUTHORIZATION:
				break;
			case SKYPECONTROLAPI_ATTACH_REFUSED:
				break;
			case SKYPECONTROLAPI_ATTACH_NOT_AVAILABLE:
				break;
			case SKYPECONTROLAPI_ATTACH_API_AVAILABLE:
				break;
			}
			lReturnCode = 1;
			if (skype->al) {
				skype->al(ulParam);
			}
			break;
		}
		fIssueDefProc = true;
		break;
	}
	if (fIssueDefProc) {
		lReturnCode = DefWindowProc(hWindow, uiMessage, uiParam, ulParam);
	}
	return (lReturnCode);
}
